-- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-- DISCLAIMER - Execute at your own risk. Read this script and thoroughly test first!!!
--
-- Written by Kimberly L. Tripp of SYSolutions, Inc.
-- For more code samples go to http://www.sqlskills.com
-- ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

SET STATISTICS IO ON
-- Turn Graphical Showplan ON (Ctrl+K)

USE CREDIT
go

-- Indexed Views v. performing the aggregate each time!
-- ONLY WORKS IN ENTERPRISE EDITION (or Developer)

--Original Query
SELECT member_no AS MemberNo, 
	sum(charge_amt) AS TotalSales
FROM dbo.charge
GROUP BY member_no
go

CREATE INDEX Covering1 on charge(charge_amt, member_no) -- Not in order by the Group By 
go

-- Run the query again - you'll see HASH AGGREGATE but on a smaller
-- set
go

SELECT member_no AS MemberNo, 
	sum(charge_amt) AS TotalSales
FROM dbo.charge
GROUP BY member_no
go

CREATE INDEX Covering2 on charge(member_no, charge_amt) -- In Order by the Group By 
go

-- Run the query again - you'll see STREAM
SELECT member_no AS MemberNo, 
	sum(charge_amt) AS TotalSales
FROM dbo.charge
GROUP BY member_no
go

-- Compare all three of these by selecting the following three and then
-- running with SHOWPLAN on.
SELECT member_no AS MemberNo, 
	sum(charge_amt) AS TotalSales
FROM dbo.charge (index = 0) -- forces a table scan
GROUP BY member_no
go

SELECT member_no AS MemberNo, 
	sum(charge_amt) AS TotalSales
FROM dbo.charge (index = Covering1) -- forces the HASH Match
GROUP BY member_no
go

SELECT member_no AS MemberNo, 
	sum(charge_amt) AS TotalSales
FROM dbo.charge (index = Covering2) -- forces the Stream 
GROUP BY member_no
go

-- So next we'll move on to Indexed View. 
-- You start with a REGULAR view - you always start with a 
-- regular view BUT with some CAVEATs here... If you want an
-- aggregate (like SUM) then you need the count of rows in the VIEW
-- For a GROUP BY in a view that will be indexed - you must use 
-- COUNT_BIG! 
-- So this is a view which is VERY close to the query but also 
-- includes the COUNT_BIG.

CREATE VIEW dbo.SumOfAllChargesByMember
	WITH SCHEMABINDING  -- required if you want indexes!
AS
SELECT member_no AS MemberNo, 
	count_big(*) AS NumberOfCharges,
	sum(charge_amt) AS TotalSales
FROM dbo.charge
GROUP BY member_no
go

-- Access the view with a SELECT...and the data needs to be 
-- computed on the fly!
SELECT * FROM dbo.SumOfAllChargesByMember
go

-- CHECK out the statistics io and the time AND turn on 
-- graphical showplan with Ctrl+K
-- DBCC FREEPROCCACHE --(to clear plans from cache)
-- DBCC DROPCLEANBUFFERS --(to clear data from cache)
SET STATISTICS IO ON
DECLARE @StartTime	datetime, @EndTime	datetime
SELECT @StartTime = getdate()
SELECT * FROM dbo.SumOfAllChargesByMember
SELECT @EndTime = getdate()
SELECT 'TOTAL TIME (ms)' = datediff (ms, @StartTime, @EndTime)
-- Keep track of time here = ????ms
-- Keep track of i/o here = ???i/o

go

-- Next CREATE the Index on the View to Materialize the data
-- fyi - Oracle has a similar feature called Materialized Views, 
-- we call them Indexed Views. They are in fact just another 
-- index on the table and will be used by any query that can 
-- benefit from them!

CREATE UNIQUE CLUSTERED INDEX SumofAllChargesIndex
	ON dbo.SumOfAllChargesByMember (MemberNo) 
		       -- notice the column name used here!
--DROP INDEX dbo.SumOfAllChargesByMember.SumofAllChargesIndex
go

-- Now RESELECT from your view and CHECK out the statistics io 
-- the time AND the graphical showplan with Ctrl+K
-- DBCC FREEPROCCACHE --(to clear plans from cache)
-- DBCC DROPCLEANBUFFERS --(to clear data from cache)
SET STATISTICS IO ON
DECLARE @StartTime	datetime, @EndTime	datetime
SELECT @StartTime = getdate()
SELECT * FROM dbo.SumOfAllChargesByMember
order by memberno
SELECT @EndTime = getdate()
SELECT 'TOTAL TIME (ms)' = datediff (ms, @StartTime, @EndTime)
-- Keep track of time here ???
-- BUT realize that Indexed Views are only accessible AND used
-- on Developer Edition and Enterprise Edition! Check your version:
SELECT @@Version
-- OR
SELECT SERVERPROPERTY('Edition')
go

-- And even better - queries the can benefit from the index WILL!
-- THIS Query will also use the indexed view (yeah!)

SELECT member_no as MemberNo, 
	avg(charge_amt) AS TotalSales
FROM dbo.charge
GROUP BY member_no
go

-- And how about an Inline function for
-- efficient filtering for the indexed view
-- Parameterized Views...
CREATE FUNCTION fn_SumforMember
	(@member_no int)
RETURNS TABLE 
AS
	RETURN SELECT *
		FROM dbo.SumOfAllChargesByMember
		WHERE MemberNo = @member_no
GO

CREATE PROC SumforMember
	(@member_no int)
AS
SELECT *
	FROM dbo.SumOfAllChargesByMember
WHERE MemberNo = @member_no
GO

-- =============================================================
-- Example to access a view, a function, and a stored procedure
-- =============================================================
set statistics io on
go
SELECT * FROM dbo.SumOfAllChargesByMember 
    WHERE memberno = 9999
	-- Same as function but more typing not as easy to use

SELECT * FROM fn_SumforMember(9999) 
	-- Can add where clauses and programmatically analyze the results

EXEC SumForMember 9999
	-- just get the data the way the developer wants you to
	-- can be more optimized and you have flexibility in recompilation
GO
sp_help charge

GO

SELECT m.Lastname, avg(charge_amt) AS TotalSales
FROM dbo.charge as c
	join member as m on m.member_no = c.member_no
GROUP BY m.lastname
go

-- Finally you can compare all 4...but a bunch of hints are needed to get
-- SQL Server to avoid using the fast/efficient index on the view...

--Here are the final comparison queries!
SET STATISTICS IO ON
go
SELECT member_no AS MemberNo, 
	sum(charge_amt) AS TotalSales
FROM dbo.charge (index = 0) -- table scan
GROUP BY member_no
OPTION (EXPAND VIEWS) -- this is to force sql server to go to the base table
go

SELECT member_no AS MemberNo, 
	sum(charge_amt) AS TotalSales
FROM dbo.charge (index = Covering1)
GROUP BY member_no
OPTION (EXPAND VIEWS)
go

SELECT member_no AS MemberNo, 
	sum(charge_amt) AS TotalSales
FROM dbo.charge (index = Covering2)
GROUP BY member_no
OPTION (EXPAND VIEWS)
go

SELECT member_no AS MemberNo, 
	sum(charge_amt) AS TotalSales
FROM dbo.charge
GROUP BY member_no
go

-- DROP ALL Indexes Added for this Demo
-- DROP INDEX Charge.Covering1
-- DROP INDEX Charge.Covering2
-- DROP INDEX dbo.SumOfAllChargesByMember.SumofAllChargesIndex

